/*--------------------------------------------------------------*
 * Morse Code to Eprom Hex File Program cw2hex Version 1.2.8	*
 * Copyright G. Forrest Cook WB0RIO - 1996-2016			*
 *								*
 * This code has been released under the			*
 * GNU GENERAL PUBLIC LICENSE version 2: http://www.gnu.org/	*
 *								*
 *  Version History:						*
 *  v1.2.2 10/25/2013 Extra characters added			*
 *  v1.2.3 7/15/2015 Fixed hex file checksums			*
 *  v1.2.4 7/15/2015 Added FSCW signal on bit 2			*
 *  v1.2.5 7/17/2015 Changed timing after letters and spaces	*
 *   to 3/7, reorganized character table and improved		*
 *   documentation.						*
 *  v1.2.6 7/20/2015 Added -c option for CSV file output and	*
 *   added support for the tx enable bit.			*
 *  v1.2.7 8/3/2015 rearranged the output bits			*
 *  v1.2.8 9/27/2016 added FSCW toggle				*
 *								*
 * An example beacon circuit that works with this program is	*
 *  here: http://www.solorb.com/elect/hamcirc/beacon/		*
 *								*
 * This program takes text input and converts it to Morse	*
 * encoded characters in the form of a standard			*
 * Intel I8HEX HEX file.					*
 * The Hex file output can be used to program an EPROM which	*
 * can then be used in a simple circuit which generates		*
 * canned CW messages.  Possible uses are CW beacon		*
 * transmitters and CQ calling machines.			*
 *								*
 * If you start the program with the -c option, it will produce	*
 * a csv output file which can be used to build an array for an	*
 * Arduino microcontroller beacon program.			*
 *								*
 * The 1989 ARRL Handbook page 19-2 documents some of the	*
 * more obscure Morse code characters, most of them are 	*
 * implemented in this program.					*
 *								*
 * The output data uses the lower four bits as follows:		*
 * The D0 output bit is the keyed morse code line.		*
 * The D1 bit has keyed Frequency Shift CW (FSCW) data,		*
 *  optionally use this to shift the frequency of dashes.	*
 * The D2 bit is the TX Enable line.				*
 * The D3 output bit is used to reset a counter in continuous	*
 * loop mode or to stop a counter in one-shot mode.		*
 *								*
 * The input string for a typical beacon message might be:	*
 * _ >> ~WB0RIO/B GSDM79IX_<<<<					*
 * The underscore starts the oscillator, the >> produces a	*
 * steady tone followed by (DE) CALLSIGN/B Grid Square Data,	*
 * the second underscore turns the oscillator off and a		*
 * silent pause ends the message.				*
 *								*
 *--------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>

#define HEXFILENAME "cw.hex"
#define CSVFILENAME "cw.csv"
#define FSCWBIT 1		/* output bit 1 */
#define FSCWMASK 0xFD		/* and mask to turn off FSCW bit */
#define TXENABLE 4		/* output bit 2 */
#define STOPBIT 8		/* output bit 3 */

/*
 * function declarations
 */
int outhexbyte();
int outcsvbyte();
void hexbout();
void hexwout();
int csumfinalize();

/*
 * Global variables
 */
FILE *Outfile=0;	/* Output file pointer */
int CSVout=0;		/* CSV (Comma Separated Values) output flag */


int main (int argc, char **argv)
{
	char *cp, uc, instr[500];
	int i, j, len, opt;
	int outval;
	int txenable = 0;
	int fscwenable = 0;
/*
 * This table contains the encoded Morse code characters.
 *
 * In each output character, bit 0 is the keying bit and bit 1 is the
 * frequency shift bit, bit 1 is high only during dashes.
 * In other words, dots are 1 and dashes are 333.
 * Inter element gaps and trailing spaces are 0, 
 * A 9 ends each character definition, it is not transmitted.
 */
	static unsigned char morse[60][24]=
	{
	  {1,0,3,3,3,0,0,0,9},			      /* A (index 0) */
	  {3,3,3,0,1,0,1,0,1,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,1,0,1,0,0,0,9},
	  {1,0,0,0,9},
	  {1,0,1,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,0,0,9},
	  {1,0,1,0,1,0,1,0,0,0,9},
	  {1,0,1,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,1,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,1,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,1,0,0,0,9},
	  {1,0,1,0,1,0,0,0,9},
	  {3,3,3,0,0,0,9},
	  {1,0,1,0,3,3,3,0,0,0,9},
	  {1,0,1,0,1,0,3,3,3,0,0,0,9},
	  {1,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,1,0,3,3,3,0,0,0,9},
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,0,0,9},
	  {3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},		/* Z (index 25) */
	  {0,0,0,0,0,0,0,9},				/* space (index 26) */
	  {3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},	/* 0 */
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,0,9},	/* 1 */
	  {1,0,1,0,3,3,3,0,3,3,3,0,3,3,3,0,0,0,9},		/* 2 */
	  {1,0,1,0,1,0,3,3,3,0,3,3,3,0,0,0,9},			/* 3 */
	  {1,0,1,0,1,0,1,0,3,3,3,0,0,0,9},			/* 4 */
	  {1,0,1,0,1,0,1,0,1,0,0,0,9},				/* 3 */
	  {3,3,3,0,1,0,1,0,1,0,1,0,0,0,9},			/* 5 */
	  {3,3,3,0,3,3,3,0,1,0,1,0,1,0,0,0,9},			/* 7 */
	  {3,3,3,0,3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},		/* 8 */
	  {3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},		/* 9 */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,3,3,3,0,0,0,9},	/* ! KW */
	  {1,0,3,3,3,0,1,0,1,0,3,3,3,0,1,0,0,0,9},		/* " AF */
	  {1,0,1,0,1,0,3,3,3,0,1,0,0,0,9},			/* ack # SN */
	  {1,0,1,0,1,0,3,3,3,0,1,0,1,0,3,3,3,0,0,0,9},		/* $ SX */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,1,0,0,0,9},	     /* parag. % AL */
	  {1,0,3,3,3,0,1,0,1,0,1,0,0,0,9},			/* & AS */
	  {1,0,3,3,3,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},	/* ' WG */
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,1,0,0,0,9},		/* ( KN */
	  {3,3,3,0,1,0,3,3,3,0,3,3,3,0,3,3,3,0,1,0,0,0,9},	/* ) KG */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},	     /* start * KA */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},			/* eom + AR */
	  {3,3,3,0,3,3,3,0,1,0,1,0,3,3,3,0,3,3,3,0,0,0,9},	/* , GW */
	  {3,3,3,0,1,0,1,0,1,0,1,0,3,3,3,0,0,0,9},		/* - DU */
	  {1,0,3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},		/* . RK */
	  {3,3,3,0,1,0,1,0,3,3,3,0,1,0,0,0,9},			/* / DN */
	  {3,3,3,0,3,3,3,0,3,3,3,0,1,0,1,0,1,0,0,0,9},		/* : OS */
	  {3,3,3,0,1,0,3,3,3,0,1,0,3,3,3,0,1,0,0,0,9},		/* ; KR */
	  {3,3,3,0,1,0,1,0,1,0,3,3,3,0,0,0,9},			/* = BT */
	  {1,0,1,0,3,3,3,0,3,3,3,0,1,0,1,0,0,0,9},		/* ? IMI */
	  {1,0,1,0,1,0,3,3,3,0,1,0,3,3,3,0,0,0,9},		/* end | SK */
	  {3,3,3,0,1,0,1,0,0,1,0,0,0,9},			/* ID ~ DE */
	  {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,9},		/* < pause */
	  {1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9},		/* > tone */
	};

	/*
	 * Print the startup message.
	 */
	printf ("\nWB0RIO CW to Intel Hex file encoder: cw2hex V1.2.8\n");
	printf ("\nuse cw2hex -c to output a cw.csv file.");
	printf ("\nThis is free software, you are welcome to");
	printf ("\nredistribute it under the restrictions of the");
	printf ("\nGNU copyright, see the GPLv2 docs for details.\n");
	printf ("\nValid Chars: A-Z, 0-9, space and special characters:\n");
	printf ("\nEnter      Send          Enter      Send");
	printf ("\n-----      ----          -----      ----");
	printf ("\n& wait     (AS)          \"          (AF)");
	printf ("\n# ack      (SN)          \'          (WG)");
	printf ("\n%% parag.   (AL)          $          (SX)");
	printf ("\n~ id       (D E)         (          (KN)");
	printf ("\n* start    (KA)          )          (KG)");
	printf ("\n+ eom      (AR)          ,          (GW)");
	printf ("\n| end      (SK)          -          (DU)");
	printf ("\n= break    (BT)          .          (RK)");
	printf ("\n:          (OS)          ;          (KR)");
	printf ("\n?          (UD)          !          (KW)");
	printf ("\n/          (DN)");
	printf ("\n^ Toggle FSCW mode       _ Toggle TX Enable");
	printf ("\n<  Pause 20 Units        >  Tone 20 Units");
	printf ("\n");

	/*
	 * Check for command line arguments.
	 */
        while ((opt = getopt (argc, argv, "c")) != EOF)
        {
                switch (opt)
                {
                /*
                 * csv output format.
                 */
                    case 'c':
			CSVout = 1;
                        break;
                /*
                 * Other stuff.
                 */
                    default:
                        printf ("Unknown option: %c\n", opt);
                        break;
                }
        }


/*
 * Open the output file.
 */
	if (CSVout)	/* csv format */
	{
		if((Outfile = fopen(CSVFILENAME, "w")) == NULL)
		{
			printf("cw2hex: can't open output file %s\n",
					CSVFILENAME);
			exit(1);
		}
	}
	else		/* hex format */
	{
		if((Outfile = fopen(HEXFILENAME, "w")) == NULL)
		{
			printf("cw2hex: can't open output file %s\n",
					HEXFILENAME);
			exit(1);
		}
	}

	cp=instr;	/* point to beginning of the input string */

/*
 * Get the string to encode here.
 */
	printf ("\nType a message to encode: ");
	fgets (instr, sizeof(instr), stdin);

/*
 * Loop through all of the characters of the input string.
 */
	while (*cp)
	{
		uc=toupper (*cp++);

/*
 * Convert letters to upper case, generate an index into the
 * Morse array for letters, punctuation and numbers.
 * The switch statement sorts out the rest of the oddball characters.
 */
		if (uc == '\n')			/* ignore trailing newline */
			break;
		else if (uc == ' ')			/* space */
			i=26;
		else if (uc >= 'A' && uc <= 'Z')	/* letters */
			i=(int)(uc - 'A');
		else if (uc >= '0' && uc <= '9')	/* numbers */
			i=(int)(uc - '0' + 27);
		else if (uc >= '!' && uc <= '/')	/* ASCII punctuation */
			i=(int)(uc - '!' + 37);
		else					/* special chars */
		{
			switch (uc)
			{
			case ':':
				i=52;
			break;
			case ';':
				i=53;
			break;
			case '=':
				i=54;
			break;
			case '?':
				i=55;
			break;
			case '|':
				i=56;
			break;
			case '~':
				i=57;
			break;
			case '<':
				i=58;
			break;
			case '>':
				i=59;
			break;
			case '_':
				if (txenable) txenable=0;
				else txenable = TXENABLE;
			break;
			case '^':
				if (fscwenable) fscwenable=0;
				else fscwenable = 1;
			break;
			default:
				i=55;  /* undefined input, output ? */
			break;
			}
		}

		printf ("\n(%c) ", uc);

		j=0;

/*
 * Process mode changes or output each byte in a given Morse character.
 */
		if (uc == '_')	/* check for Tx enable/disable */
		{
			if (txenable)
				printf ("Tx Enable On");
			else
				printf ("Tx Enable Off");
		}
		else if (uc == '^')	/* check for FSCW enable/disable */
		{
			if (fscwenable)
				printf ("FSCW On");
			else
				printf ("FSCW Off");
		}
		else
		{
		    while (morse[i][j] != 9)
		    {
			if (fscwenable)
				outval=(int)morse[i][j++]+txenable;
			else
				outval=(int)(morse[i][j++]&FSCWMASK)+txenable;

			if (CSVout)
				outcsvbyte (outval,0);
			else
				outhexbyte (outval,0);
		    }
		}
	}

	printf ("\n(stop) ");

/*
 * Get the final length of the file then close up shop.
 */
	if (CSVout)
	{
		len=outcsvbyte (STOPBIT,0);	/* output stop bit character */
		outcsvbyte (0,1);		/* finalize the csv file */
	}
	else
	{
		len=outhexbyte (STOPBIT,0);	/* output stop bit character */
		outhexbyte (0,1);		/* finalize the hex file */
	}

	fclose (Outfile);

	printf ("\n\nThis message is %d bytes long.", len);

	if (CSVout)
		printf ("\nThe output file is called %s\n\n", CSVFILENAME);
	else
		printf ("\nThe output file is called %s\n\n", HEXFILENAME);

	return (-1);
}

/*--------------------------------------------------------------*/

/*
 * Output a CSV format file.
 */

#define CLINEBYTES 10

int outcsvbyte (int byte, int finish)
{
	static int bcount=0;
	static int addr=0;

	if (!finish)	/* output a byte */
	{
		printf ("%01X", byte);

		fprintf (Outfile, "0x");	/* print one data element */
		hexbout (byte);

		if (bcount < CLINEBYTES-1)	/* not last byte in line */
		{
			fprintf (Outfile, ", ");
			bcount++;
		}
		else				/* last byte in line */
		{
			fprintf (Outfile, ",\n");
			bcount=0;
		}
		addr++;
	}
	else	/* finish */
	{
		fprintf (Outfile, "\n\n%d\n", addr);	/* Byte Count */

		return (0);
	}

	return (addr);
}

/*--------------------------------------------------------------*/

/*
 * Output bytes to an Intel Hex format file.
 */

#define HLINEBYTES 16

int outhexbyte (int byte, int finish)
{
	static int bcount=0;
	static int firstline=1;
	static int chksum=HLINEBYTES; /* first part of checksum is byte len. */
	static int addr=0;
	int i;

	if (firstline)
	{
		firstline=0;
		fprintf (Outfile, ":10000000"); /* Write beginning of file. */
	}

	if (!finish)	/* output a byte */
	{
		printf ("%01X", byte);

		if (bcount < HLINEBYTES)	/* not last byte in line */
			hexbout (byte);		/* output one hex byte */
		else	/* end of line, output checksum and start new line */
		{
			hexbout (csumfinalize(chksum));

			bcount=0;

			fprintf (Outfile, "\n:");

			hexbout (HLINEBYTES);
			chksum=HLINEBYTES;

			hexwout (addr);
			chksum += addr>>8;
			chksum += addr&255;

			fprintf (Outfile, "00");	/* pad byte */

			hexbout (byte);
		}
		bcount++;
		chksum += byte;
		addr++;
	}
	else	/* Fill out the last line with 00 bytes. */
	{
		for (i=1; i<=HLINEBYTES-bcount; i++)
		{
			hexbout (0);
			addr++;
		}

		hexbout (csumfinalize(chksum));

		fprintf (Outfile, "\n:00000001FF\n");	/* write EOF line */

		return (0);
	}

	return (addr);
}

/*--------------------------------------------------------------*/

void hexbout(byte)	/* output a byte in hex */
int     byte;
{
	char    *hexstr = { "0123456789ABCDEF" } ;

	byte = byte & 0377;
	fprintf (Outfile,"%c%c",hexstr[byte>>4],hexstr[byte&017]);
}

/*--------------------------------------------------------------*/

void hexwout(word)	/* output a word in hex */
int     word;
{
        hexbout(word>>8);	/* high byte of word */
	hexbout(word&255);	/* low byte of word */
}

/*--------------------------------------------------------------*/

int csumfinalize(dval)	/* calculate final checksum value */
int     dval;
{
	int rval;

	dval &= 255;			/* chop off incoming high bits */

	rval = (~dval + 1) & 255;	/* twos complement and mask */

	return (rval);
}

/*--------------------------------------------------------------*/
